{ -------------------------------------------------------------------------------------}
{ A "SNMP" component for Delphi32.                                                     }
{ Copyright 2000, Jean-Fabien Connault.  All Rights Reserved.                          }
{ This component can be freely used and distributed in commercial and private          }
{ environments, provided this notice is not modified in any way.                       }
{ -------------------------------------------------------------------------------------}
{ Feel free to contact me if you have any questions, comments or suggestions at        }
{   JFConnault@mail.dotcom.fr (Jean-Fabien Connault)                                   }
{ You can always find the latest version of this component at:                         }
{   http://www.worldnet.net/~cycocrew/delphi/                                          }
{ -------------------------------------------------------------------------------------}
{ -------------------------------------------------------------------------------------}
{ Date last modified:  09/17/00                                                        }
{ -------------------------------------------------------------------------------------}

{ -------------------------------------------------------------------------------------}
{ TSNMPSynapse v1.02                                                                   }
{ -------------------------------------------------------------------------------------}
{ Description:                                                                         }
{   A component that encapsulates the Synapse SNMP library                             }
{ Properties:                                                                          }
{  Public                                                                              }
{   property ErrorStatus: integer;                                                     }
{   property ErrorIndex: integer;                                                      }
{  Published                                                                           }
{   property Community: string;                                                        }
{   property Host: string;                                                             }
{   property Timeout: integer;                                                         }
{ Procedures and Functions:                                                            }
{   function SNMPGet(Oid: string): string;                                             }
{   function SNMPSet(Oid, Value: string; ValueType: TSNMPMibValueType): string;        }
{ Needs:                                                                               }
{   Synapse package from Lukas Gebauer (gebauerl@mlp.cz)                               }
{                                                                                      }
{ See example contained in example.zip file for more details.                          }
{ -------------------------------------------------------------------------------------}
{ Revision History:                                                                    }
{ 1.00:  + Initial release (08/10/00)                                                  }
{ 1.01:  + Moved some constants from SNMPSend to SNMPSynapse (08/21/00)                }
{ 1.02:  + Moved TSNMPMibValueType from SNMPSend to SNMPSynapse (09/17/00)             }
{ -------------------------------------------------------------------------------------}

unit SNMPSynapse;

{$IFNDEF WIN32}
ERROR! This unit only available on Win32!
{$ENDIF}

interface

uses
  SysUtils, Windows, Classes, StdCtrls, Forms, Dialogs, Controls, SNMPSend, ASN1Util;

type
  TSNMPMibValueType = (smvtInteger, smvtOctetString, smvtNull, smvtObjectId,
    smvtSequence, smvtIpAddress, smvtCounter, smvtGauge, smvtTimeTicks);

  TSNMPSynapse = class(TComponent)
  private
    FCommunity: string;
    FErrorStatus: integer;
    FErrorIndex: integer;
    FHost: string;
    FSNMPSend: TSNMPSend;
    FTimeout: integer;
  protected
    function ConvertValueType(ValueType: TSNMPMibValueType): integer;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function SNMPGet(Oid: string): string;
    function SNMPSet(Oid, Value: string; ValueType: TSNMPMibValueType): string;
    property ErrorStatus: integer read FErrorStatus;
    property ErrorIndex: integer read FErrorIndex;
  published
    property Community: string read FCommunity write FCommunity;
    property Host: string read FHost write FHost;
    property Timeout: integer read FTimeout write FTimeout;
  end;

procedure Register;

const
  SNMP_SUCCESS = 'Success';
  SNMP_FAILURE = 'Failure';

implementation

{***************************************************************************}
{ CONSTRUCTOR Create                                                        }
{***************************************************************************}

constructor TSNMPSynapse.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FCommunity := 'public';
  FSNMPSend := TSNMPSend.Create;
  FTimeout := 5000;
end;

{***************************************************************************}
{ DESTRUCTOR Destroy                                                        }
{***************************************************************************}

destructor TSNMPSynapse.Destroy;
begin
  FSNMPSend.Free;
  inherited Destroy;
end;

{***************************************************************************}
{ FUNCTION ConvertValueType                                                 }
{***************************************************************************}

function TSNMPSynapse.ConvertValueType(ValueType: TSNMPMibValueType): integer;
begin
  result := ASN1_NULL;
  if (ValueType = smvtInteger) then result := ASN1_INT;
  if (ValueType = smvtOctetString) then result := ASN1_OCTSTR;
  if (ValueType = smvtNull) then result := ASN1_NULL;
  if (ValueType = smvtObjectId) then result := ASN1_OBJID;
  if (ValueType = smvtSequence) then result := ASN1_SEQ;
  if (ValueType = smvtIpAddress) then result := ASN1_IPADDR;
  if (ValueType = smvtCounter) then result := ASN1_COUNTER;
  if (ValueType = smvtGauge) then result := ASN1_GAUGE;
  if (ValueType = smvtTimeTicks) then result := ASN1_TIMETICKS;
end;

{***************************************************************************}
{ FUNCTION SNMPGet                                                          }
{***************************************************************************}

function TSNMPSynapse.SNMPGet(Oid: string): string;
begin
  FSNMPSend.Query.Clear;
  FSNMPSend.Query.Community := FCommunity;
  FSNMPSend.Query.PDUType := PDUGetRequest;
  FSNMPSend.Query.MIBAdd(Oid, '', ASN1_NULL);
  FSNMPSend.Host := FHost;
  FSNMPSend.Timeout := FTimeout;
  if (FSNMPSend.DoIt = true) then
    result := FSNMPSend.Reply.MIBGet(Oid)
  else
    result := SNMP_FAILURE;
  FErrorStatus := FSNMPSend.Query.ErrorStatus;
  FErrorIndex := FSNMPSend.Query.ErrorIndex;
  inc(FSNMPSend.Query.ID);
end;

{***************************************************************************}
{ FUNCTION SNMPSet                                                          }
{***************************************************************************}

function TSNMPSynapse.SNMPSet(Oid, Value: string; ValueType: TSNMPMibValueType): string;
begin
  FSNMPSend.Query.Clear;
  FSNMPSend.Query.Community := FCommunity;
  FSNMPSend.Query.PDUType := PDUSetRequest;
  FSNMPSend.Query.MIBAdd(Oid, Value, ConvertValueType(ValueType));
  FSNMPSend.Host := FHost;
  FSNMPSend.Timeout := FTimeout;
  if (FSNMPSend.DoIt = true) then
    result := SNMP_SUCCESS
  else
    result := SNMP_FAILURE;
  FErrorStatus := FSNMPSend.Query.ErrorStatus;
  FErrorIndex := FSNMPSend.Query.ErrorIndex;
  inc(FSNMPSend.Query.ID);
end;

{***************************************************************************}
{ PROCEDURE Register                                                        }
{***************************************************************************}

procedure Register;
begin
  RegisterComponents('System', [TSNMPSynapse]);
end;

end.


